#include "audiotag.h"
#include <jack/jack.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/select.h>
#include <sndfile.h>

#define R do{printf("%s:%d:err\n",__FILE__,__LINE__);exit(1);}while(0)

const char *jack_connection[128];
int jack_connect_count;

static int process(jack_nframes_t n, void *_p)
{
  player                       *p = _p;
  jack_default_audio_sample_t  *out;
  int                          i;
  float                        *in = p->buf;
  int                          start_pos = p->buf_next;
  char                         done;

  out = jack_port_get_buffer(p->jack_port, n);

  if (GET(p->in_play) == 0) { for (i = 0; i < n; i++) out[i] = 0; return 0; }

  for (i = 0; i < n; i++) out[i] = in[(start_pos+i) & 32767];

  p->buf_next = (start_pos+n) & 32767;

  ADD(p->current, n);
  ADD(p->free_samples, n);

  done = 1;
  write(jack_process_write, &done, 1);

  return 0;
}

static void jshutdown(void *_j)
{
  fprintf(stderr, "jack shutdown\n");
  exit(1);
}

static void feed_buffer(player *p, SNDFILE *in, int channels)
{
  sf_count_t count = GET(p->free_samples);
  sf_count_t rcount;
  int i;
  int chan;
  float v;
  float *buf;

  if (!count) return;
  if (count > 32768) {
    count = 32768;
    p->free_samples = 32768;
    p->buf_next = 0;
    p->free_pos = 0;
  }

  rcount = sf_read_float(in, p->read_buf, count * channels) / channels;

  buf = p->read_buf;
  for (i = 0; i < rcount; i++) {
    v = 0;
    for (chan = 0; chan < channels; chan++, buf++)
      v += *buf;
    v /= channels;
    p->buf[(p->free_pos + i) & 32767] = v;
  }
  for (; i < count; i++)
    p->buf[(p->free_pos + i) & 32767] = 0;

  ADD(p->free_samples, -count);
  p->free_pos += count;
  p->free_pos &= 32767;
}

static void *play_thread(void *_p)
{
  player *p = _p;
  SNDFILE *in;
  SF_INFO inf;
  sf_count_t c;
  char pc;

  in = open_audio_file(p->audio_file, &inf);

  p->read_buf = malloc(sizeof(float) * 32768 * inf.channels);
  if (p->read_buf == NULL) abort();

next:
  while (1) {
    if (read(jack_player_read, &pc, 1) != 1) abort();
    if (pc == 1) break;
  }

  p->buf_next = 0;
  p->free_samples = 32768;
  p->free_pos = 0;
  SET(p->current, p->start);

  c = sf_seek(in, p->start, SEEK_SET);
  if (c != p->start) {
    fprintf(stderr, "error seeking %s at %ld\n", p->audio_file, p->start);
    goto next;
  }

  feed_buffer(p, in, inf.channels);

  SET(p->in_play, 1);
  com_jack_player_started();

  while (1) {
    fd_set r;
    FD_ZERO(&r);
    FD_SET(jack_process_read, &r);
    FD_SET(jack_player_read, &r);
    if (select(jack_process_read > jack_player_read ?
               jack_process_read + 1 : jack_player_read + 1,
               &r, NULL, NULL, NULL) == -1) abort();
    if (FD_ISSET(jack_process_read, &r)) {
      if (read(jack_process_read, &pc, 1) != 1) abort();
      feed_buffer(p, in, inf.channels);
      com_jack_player_advance();
    }
    if (FD_ISSET(jack_player_read, &r)) {
      if (read(jack_player_read, &pc, 1) != 1) abort();
      if (pc == 0) break;
    }
  }

  SET(p->in_play, 0);
  com_jack_player_ended();

  goto next;
}

void init_jack(player *pl)
{
  jack_client_t  *client;
  jack_status_t  status;
  jack_port_t    *port;
  pthread_attr_t attr;
  pthread_t      t;
  int i;

  pl->in_play = 0;
  if (pl->no_jack) return;

  if (pthread_attr_init(&attr)) abort();
  if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)) abort();
  if (pthread_create(&t, &attr, play_thread, pl)) abort();
  if (pthread_attr_destroy(&attr)) abort();

  pl->jack_error = 0;

  client = jack_client_open("audiotag", JackNoStartServer, &status);
  if (!client) {
    fprintf(stderr, "error calling jack_client_open, jack not activated\n");
    pl->jack_error = 1;
    return;
  }
  jack_set_process_callback(client, process, pl);
  jack_on_shutdown(client, jshutdown, pl);
  port = jack_port_register(client, "out", JACK_DEFAULT_AUDIO_TYPE,
                            JackPortIsOutput, 0);
  if (!port) R;
  pl->jack_port = port;
  if (jack_activate(client)) R;
  for (i = 0; i < jack_connect_count; i++)
    jack_connect(client, "audiotag:out", jack_connection[i]);
}

void start_play(player *p)
{
  if (p->no_jack) return;
  if (p->jack_error) return;
  com_jack_player_start();
}

void stop_play(player *p)
{
  if (p->no_jack) return;
  if (p->jack_error) return;
  com_jack_player_stop();
}
